#include <pty.h>
#include <stdio.h>
#include <signal.h>
#include <typeinfo>

#include <sys/stropts.h>
#include <termios.h>
#include <unistd.h>
#include <fcntl.h>

#include "idlemgr.hh"
#include "window.hh"
#include "kbd.hh"
#include "term.hh"
#include "sizes"

class termwindow : public window
{
	int fd;
	int cx, cy;
	int top, bottom;
	int intensity, underline, blink, reverse, fgc, bgc;
	BYTE attr;
	enum { ESnormal, ESesc, ESsquare, ESgetpars, ESgotpars,
           EShash, ESignore, ESescnext } state;
    vector<unsigned> par;
    bool ques;
    void ResetAttr()
    {
    	intensity = underline = blink = reverse = bgc = 0;
    	fgc = 7;
    }
    void BuildAttr()
    {
    	attr = (bgc<<4) | fgc;
    	if(intensity==-1)attr = (attr&0xF0) | 8;
    	if(reverse)attr = (attr & 0x88) | (((attr >> 4) | (attr << 4)) & 0x77);
    	if(intensity==1)attr |= 8;
    	if(blink)attr |= 0x80;
    }
    void Reset()
    {
    	state = ESnormal;
    	cx = cy = top = 0;
    	bottom = height();
    	ques = false;
    	ResetAttr();
    	BuildAttr();
    }
    void FixCoord()
    {
    	if(bottom>height())bottom=height();
    	if(top>bottom-2)top=bottom-2;
    	if(top<0)top=0;
    	if(bottom<top+2)bottom=top+2;
    	if(cx<0)cx=0; if(cx>=width())cx=width()-1;
    	if(cy<0)cy=0; if(cy>=bottom-top)cy=bottom-top-1;
    }
    void ScrollFix()
    {
		if(cx >= width()) { cx = 0; Lf(); }
    }
    void fillbox(unsigned x1,unsigned y1, unsigned x2,unsigned y2, char c=' ')
    {
    	unsigned wid = x2-x1+1;
    	for(unsigned y=y1; y<=y2; ++y)
    	{
    		winchar *line = &buf.buf[y][x1];
    		for(unsigned x=0; x<wid; ++x)
    		{
    			line[x].ch = ' ';
    			line[x].attr = attr;
    		}
    	}
    }
    void yscroll(unsigned x1,unsigned y1, unsigned x2,unsigned y2, int amount)
    {
    	unsigned hei = y2-y1+1;
    	unsigned wid = x2-x1+1;
    	if(amount > 0) /* down */
    	{
    		if(amount > hei)amount = hei;
	    	for(unsigned y=y2; y>=y1+amount; --y)
	    		buf.buf[y] = buf.buf[y-amount];
	    	fillbox(x1,y1, x2,y1+amount-1);
    	}
    	else	/*  up */
    	{
    		amount = -amount;
    		if(amount > hei)amount = hei;
	    	for(unsigned y=y1; y<=y2-amount; ++y)
	    		buf.buf[y] = buf.buf[y+amount];
	    	fillbox(x1,y2-amount+1, x2,y2);
    	}
    }
    void csi_at(unsigned c)
    {
    	// insert c spaces at cursor
    	if(!c)c=1;else if(c > width()-cx)c = width()-cx;
    	winchar *line = &buf.buf[cy][0];
    	for(unsigned a=width(); --a>cx+c; )
    		line[a] = line[a-c];
    	fillbox(cx, cy, cx+c-1, cy);
    }
    void csi_X(unsigned c)
    {
    	// write c spaces at cursor (overwrite)
    	fillbox(cx,cy+top, cx+c-1, cy+top);
    }
    void csi_P(unsigned c)
    {
    	// insert c black holes at cursor (eat c characters
    	// and scroll line horizontally to left)
    	if(!c)c=1; else if(c > width()-cx)c = width()-cx;
    	winchar *line = &buf.buf[cy][0];
    	for(unsigned a=cx; a<width()-c; ++a)
    		line[a] = line[a+c];
    	fillbox(width()-c-2, cy, width()-1, cy);
    }
    void csi_J(unsigned c)
    {
    	switch(c)
    	{
    		case 0: // erase from cursor to end of display
    			fillbox(0,cy, width()-1,height()-1);
    			break;
    		case 1: // erase from start to cursor
    			fillbox(0,0, width()-1,cy-1);
    			break;
    		case 2: // erase whole display
    			fillbox(0,0, width()-1,height()-1);
    			break;
    	}
    }
    void csi_K(unsigned c)
    {
    	switch(c)
    	{
    		case 0: // erase from cursor to end of line
    			fillbox(cx,cy, width()-1,cy);
    			break;
    		case 1: // erase from start of line to cursor
    			fillbox(0,cy, cx-1,cy);
    			break;
    		case 2: // erase whole line
    			fillbox(0,cy, width()-1,cy);
    			break;
    	}
    }
    void csi_L(unsigned c)
    {
    	// scroll the rest of window c lines down,
    	// including where cursor is. Don't move cursor.
    	yscroll(0,cy, width()-1,bottom-1, 1);
    }
    void csi_M(unsigned c)
    {
    	// do c antilinefeeds (eat c lines), scrolling
    	// the rest of the window up
    	yscroll(0,cy, width()-1,bottom-1, -1);
    }
    void Lf()
    {
    	if(++cy<bottom-top)return;
    	--cy;
    	/* scroll the window up */
    	yscroll(0,top, width()-1,bottom-1, -1);
    }
    void Ri()
    {
    	if(--cy>=0)return;
    	++cy;
    	/* scroll the window down */
    	yscroll(0,top, width()-1,bottom-1, 1);
    }
public:
    void Write(const string &s)
	{
		// This is an indirect ripoff from
		// /usr/src/linux/drivers/char/console.c
		unsigned a, b=s.size();
		for(a=0; a<b; ++a)
		{
			int c = s[a];
			switch(c)
			{
				case 0: break;
				case 7: write(1,"\7",1); /* bell */ break;
				case 8: ScrollFix(); if(cx>0)--cx; break;
				case 9: ScrollFix(); cx += 8 - (cx & 7); FixCoord(); break;
				case 10: case 11: case 12: ScrollFix(); Lf(); break;
				case 13: cx=0; break;
				case 14: case 15: /* charset */ break;
				case 24: case 26: state = ESnormal; break;
				case 27: state = ESescnext; break;
				case 127: /* del - ignore */ break;
				case 128+27: state = ESesc; break;
			}
			switch(state)
			{
				case ESescnext:
					state = ESesc;
					break;
				case ESnormal:
					if(c >= 32)
					{
						ScrollFix();
						buf.buf[cy+top][cx].attr = attr;
						buf.buf[cy+top][cx].ch = c;
						++cx;
					}
					break;
				case ESesc:
					state = ESnormal;
					switch(c)
					{
						case '[': state = ESsquare; break;
						case '%': state = ESignore; break;
						case 'E': cx=0; Lf(); break;
						case 'M': Ri(); break;
						case 'D': Lf(); break;
						case 'Z': write(fd, "\33[?6c", 5); break;
						case '(': state = ESignore; break;
						case ')': state = ESignore; break;
						case '#': state = EShash; break;
						case '7': /* save */ break;
						case '8': /* restore */ break;
						case 'c': Reset(); break;
					}
					break;
				case ESsquare:
					par.clear(); par.push_back(0);
					state = ESgetpars;
					if(c=='[') { state = ESignore; break; }
					if(c=='?') { ques=true; break; }
					ques = false;
					/* fallthru */
				case ESgetpars:
					if(c==';') { par.push_back(0); break; }
					if(c>='0' && c<='9') { par[par.size()-1] = par[par.size()-1]*10 + c-'0'; break; }
					state = ESgotpars;
					/* fallthru */
				case ESgotpars:
					state = ESnormal;
					if(ques && (c=='c' /* cursor type */
					         || c=='m' /* complement mask */))break;
					         
					switch(c)
					{
						case 'h': /* misc modes on */ break;
						case 'l': /* misc modes off */ break;
						case 'n':
							if(!ques)
							{
								if(par[0]==5)write(fd, "\033[0n", 4);
								else if(par[0]==6){char Buf[64];sprintf(Buf, "\33[%u;%uR",cy+1,cx+1);write(fd,Buf,strlen(Buf));}
							}
							break;
						case 'G': case '`': if(par[0])--par[0]; cx=par[0]; FixCoord(); break;
						case 'A': if(!par[0])par[0]=1; cy-=par[0]; FixCoord(); break;
						case 'B': if(!par[0])par[0]=1; cy+=par[0]; FixCoord(); break;
						case 'C': ScrollFix(); if(!par[0])par[0]=1; cx+=par[0]; FixCoord(); break;
						case 'D': ScrollFix(); if(!par[0])par[0]=1; cx-=par[0]; FixCoord(); break;
						case 'E': if(!par[0])par[0]=1; cx=0; cy+=par[0]; FixCoord(); break;
						case 'F': if(!par[0])par[0]=1; cx=0; cy-=par[0]; FixCoord(); break;
						case 'H': case 'f': par.resize(2);
							if(par[0])--par[0];
							if(par[1])--par[1];
							cx=par[1]; cy=par[0];
							break;
						case 'J': csi_J(par[0]); break;
						case 'K': csi_K(par[0]); break;
						case 'L': csi_L(par[0]); break;
						case 'M': csi_M(par[0]); break;
						case 'P': csi_P(par[0]); break;
						case 'X': csi_X(par[0]); break;
						case '@': csi_at(par[0]); break;
						case 'c': if(!par[0])write(fd, "\33[?6c", 5); break;
						case 'g': /* set tab stops */ break;
						case 'q': /* set leds */ break;
						case 'm':
							for(unsigned a=0; a<par.size(); ++a)
								switch(par[a])
								{
									case 0: ResetAttr(); break;
									case 1: intensity = 1; break;
									case 2: intensity = -1; break;
									case 4: underline = 1; break;
									case 5: blink = 1; break;
									case 7: reverse = 1; break;
									case 21: case 22: intensity = 0; break;
									case 24: underline = 0; break;
									case 25: blink = 0; break;
									case 27: reverse = 0; break;
									default:
									{
										static const BYTE swap[8] = {0,4,2,6,1,5,3,7};
										if(par[a]>=30 && par[a]<=37)fgc = swap[par[a]-30];
										else if(par[a]>=40 && par[a]<=47)bgc = swap[par[a]-40];
									}
								}
							BuildAttr();
							break;
						case 'r': par.resize(2);
							if(!par[0])par[0]=1;
							if(!par[1])par[1]=height();
							if(par[0]<par[1] && par[1]<=height())
							{
								top=par[0]-1, bottom=par[1]-1;
								cx=0, cy=0;
							}
							break;
						case 's': /* save */ break;
						case 'u': /* restore */ break;
					}
					break;
				case EShash:
					state = ESnormal;
					if(c == '8') { /* clear screen with 'E' */ break; }
					break;
				case ESignore:
					state = ESnormal;
					break;
			}
		}
	}
	termwindow() : window()
	{
		Reset();
	}
	termwindow(unsigned x, unsigned y) : window(x, y)
	{
		Reset();
  	}
	virtual void resize(unsigned nxs, unsigned nys)
	{
		window::resize(nxs, nys);
		Reset();
	}
};

class shellwindow : public termwindow
{
	int fd;
	int pid;

	void TermOpen(unsigned x, unsigned y)
	{
		struct winsize ws;
		memset(&ws, 0, sizeof ws);
		ws.ws_row = y;
		ws.ws_col = x;
		
		pid = forkpty(&fd, NULL, NULL, &ws);
		if(pid < 0)
			throw "forkpty failed - maybe we need rights to /dev/ptmx?";
		
		if(!pid)
		{
			execl(getenv("SHELL"), getenv("SHELL"), NULL);
			/* not reached */
		}
		fcntl(fd, F_SETFL, fcntl(fd, F_GETFL) | O_NONBLOCK);

		idlemgr.registertask(this, 20);
		
		Write(
			"\33[30;47m\33[2J"
			"\33[34mKeys:\t\33[30malt+x = exit\r\n"
			"\tF5+arrows+enter = move,\r\n"
			"\tShiftF5+arrows+enter = resize\r\n-----------------------------------------\r\n"
			"\33[31mPs: The hole in the window is intentional\r\n"
			"for demo purposes. And yes, I know the cursor\r\n"
			"does not work as wanted...\33[30m\r\n");
			
		for(unsigned yy=y*4/6; yy<y*5/6; ++yy)
			for(unsigned xx=x*4/6; xx<x*5/6; ++xx)
				buf.buf[yy][xx].set=false;
	}
	void TermClose()
	{
		idlemgr.unregistertask(this);
		
		kill(pid, SIGKILL);
		close(fd);
	}
public:
	shellwindow(unsigned x, unsigned y) : termwindow(x, y)
	{
		TermOpen(x, y);
  	}
  	virtual ~shellwindow()
  	{
  		TermClose();
  	}
	virtual void resize(unsigned nxs, unsigned nys)
	{
		termwindow::resize(nxs, nys);
		
		struct winsize ws;
		memset(&ws, 0, sizeof ws);
	    ws.ws_col = nxs;
	    ws.ws_row = nys;
	    ioctl(fd, TIOCSWINSZ, &ws);
	}
	virtual void handleevent(event &e)
	{
		window::handleevent(e);
		if(typeid(*e) == typeid(event_idle))
		{
			char Buf[2048];
			int c = read(fd, Buf, sizeof Buf);
			if(c > 0)
				Write(string(Buf, c));
		}

        if(event_key *p = dynamic_cast<event_key *> (&*e))
        {
        	write(fd, p->ch.data(), p->ch.size());
        }
	}
};

#include "demo.hh"


int main()
{
	TermInit();
	try
	{
		desktop->addchild(new demowin(70,22), 1,1);
		desktop->addchild(new shellwindow(50,19), 5,2);
	}
	catch(const char *s)
	{
		fprintf(stderr, "%s\r\n", s);
		return -1;
	}
	TermLoop();
	TermDone();
	
	return 0;
}
